@@ -45,18 +45,10 @@ typedef enum {
 #define HB_SIZE (1 << HB_BITS)
 #define DFLTCC_FACILITY 151
 
-local inline dfltcc_cc dfltcc OF((int fn, void *param,
-                                  Bytef **op1, size_t *len1,
-                                  z_const Bytef **op2, size_t *len2,
-                                  void *hist));
-local inline dfltcc_cc dfltcc(fn, param, op1, len1, op2, len2, hist)
-    int fn;
-    void *param;
-    Bytef **op1;
-    size_t *len1;
-    z_const Bytef **op2;
-    size_t *len2;
-    void *hist;
+local inline dfltcc_cc dfltcc(int fn, void *param,
+			      Bytef **op1, size_t *len1,
+			      z_const Bytef **op2, size_t *len2,
+			      void *hist)
 {
     Bytef *t2 = op1 ? *op1 : NULL;
     size_t t3 = len1 ? *len1 : 0;
@@ -123,18 +115,12 @@ struct dfltcc_qaf_param {
 static_assert(sizeof(struct dfltcc_qaf_param) == 32,
               sizeof_struct_dfltcc_qaf_param_is_32);
 
-local inline int is_bit_set OF((const char *bits, int n));
-local inline int is_bit_set(bits, n)
-    const char *bits;
-    int n;
+local inline int is_bit_set(const char *bits, int n)
 {
     return bits[n / 8] & (1 << (7 - (n % 8)));
 }
 
-local inline void clear_bit OF((char *bits, int n));
-local inline void clear_bit(bits, n)
-    char *bits;
-    int n;
+local inline void clear_bit(char *bits, int n)
 {
     bits[n / 8] &= ~(1 << (7 - (n % 8)));
 }
@@ -197,10 +183,7 @@ struct dfltcc_param_v0 {
 static_assert(sizeof(struct dfltcc_param_v0) == 1536,
               sizeof_struct_dfltcc_param_v0_is_1536);
 
-local z_const char *oesc_msg OF((char *buf, int oesc));
-local z_const char *oesc_msg(buf, oesc)
-    char *buf;
-    int oesc;
+local z_const char *oesc_msg(char *buf, int oesc)
 {
     if (oesc == 0x00)
         return NULL; /* Successful completion */
@@ -226,27 +209,19 @@ struct dfltcc_state {
 #define ALIGN_UP(p, size) \
         (__typeof__(p))(((uintptr_t)(p) + ((size) - 1)) & ~((size) - 1))
 
-#define GET_DFLTCC_STATE(state) ((struct dfltcc_state FAR *)( \
-        (char FAR *)(state) + ALIGN_UP(sizeof(*state), 8)))
+#define GET_DFLTCC_STATE(state) ((struct dfltcc_state *)( \
+        (char *)(state) + ALIGN_UP(sizeof(*state), 8)))
 
 /*
    Compress.
  */
 local inline int dfltcc_can_deflate_with_params(z_streamp strm,
-                                                int level,
-                                                uInt window_bits,
-                                                int strategy);
-local inline int dfltcc_can_deflate_with_params(strm,
-                                                level,
-                                                window_bits,
-                                                strategy)
-    z_streamp strm;
-    int level;
-    uInt window_bits;
-    int strategy;
+						int level,
+						uInt window_bits,
+						int strategy)
 {
-    deflate_state FAR *state = (deflate_state FAR *)strm->state;
-    struct dfltcc_state FAR *dfltcc_state = GET_DFLTCC_STATE(state);
+    deflate_state *state = (deflate_state *)strm->state;
+    struct dfltcc_state *dfltcc_state = GET_DFLTCC_STATE(state);
 
     /* Unsupported compression settings */
     if ((dfltcc_state->level_mask & (1 << level)) == 0)
@@ -265,10 +240,9 @@ local inline int dfltcc_can_deflate_with_params(strm,
     return 1;
 }
 
-int ZLIB_INTERNAL dfltcc_can_deflate(strm)
-    z_streamp strm;
+int ZLIB_INTERNAL dfltcc_can_deflate(z_streamp strm)
 {
-    deflate_state FAR *state = (deflate_state FAR *)strm->state;
+    deflate_state *state = (deflate_state *)strm->state;
 
     return dfltcc_can_deflate_with_params(strm,
                                           state->level,
@@ -276,12 +250,10 @@ int ZLIB_INTERNAL dfltcc_can_deflate(strm)
                                           state->strategy);
 }
 
-local void dfltcc_gdht OF((z_streamp strm));
-local void dfltcc_gdht(strm)
-    z_streamp strm;
+local void dfltcc_gdht(z_streamp strm)
 {
-    deflate_state FAR *state = (deflate_state FAR *)strm->state;
-    struct dfltcc_param_v0 FAR *param = &GET_DFLTCC_STATE(state)->param;
+    deflate_state *state = (deflate_state *)strm->state;
+    struct dfltcc_param_v0 *param = &GET_DFLTCC_STATE(state)->param;
     size_t avail_in = avail_in = strm->avail_in;
 
     dfltcc(DFLTCC_GDHT,
@@ -289,12 +261,10 @@ local void dfltcc_gdht(strm)
            &strm->next_in, &avail_in, NULL);
 }
 
-local dfltcc_cc dfltcc_cmpr OF((z_streamp strm));
-local dfltcc_cc dfltcc_cmpr(strm)
-    z_streamp strm;
+local dfltcc_cc dfltcc_cmpr(z_streamp strm)
 {
-    deflate_state FAR *state = (deflate_state FAR *)strm->state;
-    struct dfltcc_param_v0 FAR *param = &GET_DFLTCC_STATE(state)->param;
+    deflate_state *state = (deflate_state *)strm->state;
+    struct dfltcc_param_v0 *param = &GET_DFLTCC_STATE(state)->param;
     size_t avail_in = strm->avail_in;
     size_t avail_out = strm->avail_out;
     dfltcc_cc cc;
@@ -309,13 +279,10 @@ local dfltcc_cc dfltcc_cmpr(strm)
     return cc;
 }
 
-local void send_eobs OF((z_streamp strm,
-                         z_const struct dfltcc_param_v0 FAR *param));
-local void send_eobs(strm, param)
-    z_streamp strm;
-    z_const struct dfltcc_param_v0 FAR *param;
+local void send_eobs(z_streamp strm,
+		     z_const struct dfltcc_param_v0 *param)
 {
-    deflate_state FAR *state = (deflate_state FAR *)strm->state;
+    deflate_state *state = (deflate_state *)strm->state;
 
     _tr_send_bits(
           state,
@@ -337,14 +304,12 @@ local void send_eobs(strm, param)
 #endif
 }
 
-int ZLIB_INTERNAL dfltcc_deflate(strm, flush, result)
-    z_streamp strm;
-    int flush;
-    block_state *result;
+int ZLIB_INTERNAL dfltcc_deflate(z_streamp strm, int flush,
+				 block_state *result)
 {
-    deflate_state FAR *state = (deflate_state FAR *)strm->state;
-    struct dfltcc_state FAR *dfltcc_state = GET_DFLTCC_STATE(state);
-    struct dfltcc_param_v0 FAR *param = &dfltcc_state->param;
+    deflate_state *state = (deflate_state *)strm->state;
+    struct dfltcc_state *dfltcc_state = GET_DFLTCC_STATE(state);
+    struct dfltcc_param_v0 *param = &dfltcc_state->param;
     uInt masked_avail_in;
     dfltcc_cc cc;
     int need_empty_block;
@@ -457,7 +422,10 @@ again:
         *strm->next_out = (Bytef)state->bi_buf;
     /* Honor history and check value */
     param->nt = 0;
-    param->cv = state->wrap == 2 ? ZSWAP32(strm->adler) : strm->adler;
+    if (state->wrap == 1)
+        param->cv = strm->adler;
+    else if (state->wrap == 2)
+        param->cv = ZSWAP32(strm->adler);
 
     /* When opening a block, choose a Huffman-Table Type */
     if (!param->bcf) {
@@ -489,7 +457,10 @@ again:
         state->bi_buf = 0; /* Avoid accessing next_out */
     else
         state->bi_buf = *strm->next_out & ((1 << state->bi_valid) - 1);
-    strm->adler = state->wrap == 2 ? ZSWAP32(param->cv) : param->cv;
+    if (state->wrap == 1)
+        strm->adler = param->cv;
+    else if (state->wrap == 2)
+        strm->adler = ZSWAP32(param->cv);
 
     /* Unmask the input data */
     strm->avail_in += masked_avail_in;
@@ -534,27 +505,20 @@ again:
 /*
    Expand.
  */
-int ZLIB_INTERNAL dfltcc_can_inflate(strm)
-    z_streamp strm;
+int ZLIB_INTERNAL dfltcc_can_inflate(z_streamp strm)
 {
-    struct inflate_state FAR *state = (struct inflate_state FAR *)strm->state;
-    struct dfltcc_state FAR *dfltcc_state = GET_DFLTCC_STATE(state);
-
-    /* Unsupported compression settings */
-    if (state->wbits != HB_BITS)
-        return 0;
+    struct inflate_state *state = (struct inflate_state *)strm->state;
+    struct dfltcc_state *dfltcc_state = GET_DFLTCC_STATE(state);
 
     /* Unsupported hardware */
     return is_bit_set(dfltcc_state->af.fns, DFLTCC_XPND) &&
                is_bit_set(dfltcc_state->af.fmts, DFLTCC_FMT0);
 }
 
-local dfltcc_cc dfltcc_xpnd OF((z_streamp strm));
-local dfltcc_cc dfltcc_xpnd(strm)
-    z_streamp strm;
+local dfltcc_cc dfltcc_xpnd(z_streamp strm)
 {
-    struct inflate_state FAR *state = (struct inflate_state FAR *)strm->state;
-    struct dfltcc_param_v0 FAR *param = &GET_DFLTCC_STATE(state)->param;
+    struct inflate_state *state = (struct inflate_state *)strm->state;
+    struct dfltcc_param_v0 *param = &GET_DFLTCC_STATE(state)->param;
     size_t avail_in = strm->avail_in;
     size_t avail_out = strm->avail_out;
     dfltcc_cc cc;
@@ -567,14 +531,12 @@ local dfltcc_cc dfltcc_xpnd(strm)
     return cc;
 }
 
-dfltcc_inflate_action ZLIB_INTERNAL dfltcc_inflate(strm, flush, ret)
-    z_streamp strm;
-    int flush;
-    int *ret;
+dfltcc_inflate_action ZLIB_INTERNAL dfltcc_inflate(z_streamp strm, int flush,
+						   int *ret)
 {
-    struct inflate_state FAR *state = (struct inflate_state FAR *)strm->state;
-    struct dfltcc_state FAR *dfltcc_state = GET_DFLTCC_STATE(state);
-    struct dfltcc_param_v0 FAR *param = &dfltcc_state->param;
+    struct inflate_state *state = (struct inflate_state *)strm->state;
+    struct dfltcc_state *dfltcc_state = GET_DFLTCC_STATE(state);
+    struct dfltcc_param_v0 *param = &dfltcc_state->param;
     dfltcc_cc cc;
 
     if (flush == Z_BLOCK || flush == Z_TREES) {
@@ -605,13 +567,12 @@ dfltcc_inflate_action ZLIB_INTERNAL dfltcc_inflate(strm, flush, ret)
     }
 
     /* Translate stream to parameter block */
-    param->cvt = state->flags ? CVT_CRC32 : CVT_ADLER32;
+    param->cvt = ((state->wrap & 4) && state->flags) ? CVT_CRC32 : CVT_ADLER32;
     param->sbb = state->bits;
-    param->hl = state->whave; /* Software and hardware history formats match */
-    param->ho = (state->wnext - state->whave) & ((1 << HB_BITS) - 1);
     if (param->hl)
         param->nt = 0; /* Honor history for the first block */
-    param->cv = state->flags ? ZSWAP32(state->check) : state->check;
+    if (state->wrap & 4)
+        param->cv = state->flags ? ZSWAP32(state->check) : state->check;
 
     /* Inflate */
     do {
@@ -622,9 +583,9 @@ dfltcc_inflate_action ZLIB_INTERNAL dfltcc_inflate(strm, flush, ret)
     strm->msg = oesc_msg(dfltcc_state->msg, param->oesc);
     state->last = cc == DFLTCC_CC_OK;
     state->bits = param->sbb;
-    state->whave = param->hl;
-    state->wnext = (param->ho + param->hl) & ((1 << HB_BITS) - 1);
-    state->check = state->flags ? ZSWAP32(param->cv) : param->cv;
+    if (state->wrap & 4)
+        strm->adler = state->check = state->flags ?
+                                     ZSWAP32(param->cv) : param->cv;
     if (cc == DFLTCC_CC_OP2_CORRUPT && param->oesc != 0) {
         /* Report an error if stream is corrupted */
         state->mode = BAD;
@@ -636,20 +597,55 @@ dfltcc_inflate_action ZLIB_INTERNAL dfltcc_inflate(strm, flush, ret)
         DFLTCC_INFLATE_BREAK : DFLTCC_INFLATE_CONTINUE;
 }
 
-int ZLIB_INTERNAL dfltcc_was_inflate_used(strm)
-    z_streamp strm;
+int ZLIB_INTERNAL dfltcc_was_inflate_used(z_streamp strm)
 {
-    struct inflate_state FAR *state = (struct inflate_state FAR *)strm->state;
-    struct dfltcc_param_v0 FAR *param = &GET_DFLTCC_STATE(state)->param;
+    struct inflate_state *state = (struct inflate_state *)strm->state;
+    struct dfltcc_param_v0 *param = &GET_DFLTCC_STATE(state)->param;
 
     return !param->nt;
 }
 
-int ZLIB_INTERNAL dfltcc_inflate_disable(strm)
-    z_streamp strm;
+/*
+   Rotates a circular buffer.
+   The implementation is based on https://cplusplus.com/reference/algorithm/rotate/
+ */
+local void rotate(Bytef *start, Bytef *pivot, Bytef *end)
 {
-    struct inflate_state FAR *state = (struct inflate_state FAR *)strm->state;
-    struct dfltcc_state FAR *dfltcc_state = GET_DFLTCC_STATE(state);
+    Bytef *p = pivot;
+    Bytef tmp;
+
+    while (p != start) {
+        tmp = *start;
+        *start = *p;
+        *p = tmp;
+
+        start++;
+        p++;
+
+        if (p == end)
+            p = pivot;
+        else if (start == pivot)
+            pivot = p;
+    }
+}
+
+#define MIN(x, y) ({    \
+    typeof(x) _x = (x); \
+    typeof(y) _y = (y); \
+    _x < _y ? _x : _y;  \
+})
+
+#define MAX(x, y) ({    \
+    typeof(x) _x = (x); \
+    typeof(y) _y = (y); \
+    _x > _y ? _x : _y;  \
+})
+
+int ZLIB_INTERNAL dfltcc_inflate_disable(z_streamp strm)
+{
+    struct inflate_state *state = (struct inflate_state *)strm->state;
+    struct dfltcc_state *dfltcc_state = GET_DFLTCC_STATE(state);
+    struct dfltcc_param_v0 *param = &dfltcc_state->param;
 
     if (!dfltcc_can_inflate(strm))
         return 0;
@@ -661,6 +657,9 @@ int ZLIB_INTERNAL dfltcc_inflate_disable(strm)
         return 1;
     /* DFLTCC was not used yet - decompress in software */
     memset(&dfltcc_state->af, 0, sizeof(dfltcc_state->af));
+    /* Convert the window from the hardware to the software format */
+    rotate(state->window, state->window + param->ho, state->window + HB_SIZE);
+    state->whave = state->wnext = MIN(param->hl, state->wsize);
     return 0;
 }
 
@@ -674,7 +673,6 @@ local unsigned long env_ribm;
 local uint64_t cpu_facilities[(DFLTCC_FACILITY / 64) + 1];
 local struct dfltcc_qaf_param cpu_af __attribute__((aligned(8)));
 
-local inline int is_dfltcc_enabled OF((void));
 local inline int is_dfltcc_enabled(void)
 {
     if (env_dfltcc_disabled)
@@ -684,10 +682,7 @@ local inline int is_dfltcc_enabled(void)
     return is_bit_set((const char *)cpu_facilities, DFLTCC_FACILITY);
 }
 
-local unsigned long xstrtoul OF((const char *s, unsigned long _default));
-local unsigned long xstrtoul(s, _default)
-    const char *s;
-    unsigned long _default;
+local unsigned long xstrtoul(const char *s, unsigned long _default)
 {
     char *endptr;
     unsigned long result;
@@ -699,10 +694,8 @@ local unsigned long xstrtoul(s, _default)
     return (errno || *endptr) ? _default : result;
 }
 
-__attribute__((constructor)) local void init_globals OF((void));
 __attribute__((constructor)) local void init_globals(void)
 {
-    const char *endptr;
     const char *env;
     register char r0 __asm__("r0");
 
@@ -776,12 +769,10 @@ __attribute__((constructor)) local void init_globals(void)
    `posix_memalign' is not an option. Thus, we overallocate and take the
    aligned portion of the buffer.
 */
-void ZLIB_INTERNAL dfltcc_reset(strm, size)
-    z_streamp strm;
-    uInt size;
+void ZLIB_INTERNAL dfltcc_reset(z_streamp strm, uInt size)
 {
     struct dfltcc_state *dfltcc_state =
-        (struct dfltcc_state *)((char FAR *)strm->state + ALIGN_UP(size, 8));
+        (struct dfltcc_state *)((char *)strm->state + ALIGN_UP(size, 8));
 
     memcpy(&dfltcc_state->af, &cpu_af, sizeof(dfltcc_state->af));
 
@@ -803,51 +794,45 @@ void ZLIB_INTERNAL dfltcc_reset(strm, size)
     dfltcc_state->param.ribm = env_ribm;
 }
 
-voidpf ZLIB_INTERNAL dfltcc_alloc_state(strm, items, size)
-    z_streamp strm;
-    uInt items;
-    uInt size;
+voidpf ZLIB_INTERNAL dfltcc_alloc_state(z_streamp strm, uInt items, uInt size)
 {
     return ZALLOC(strm,
                   ALIGN_UP(items * size, 8) + sizeof(struct dfltcc_state),
                   sizeof(unsigned char));
 }
 
-void ZLIB_INTERNAL dfltcc_copy_state(dst, src, size)
-    voidpf dst;
-    const voidpf src;
-    uInt size;
+void ZLIB_INTERNAL dfltcc_copy_state(voidpf dst, const voidpf src, uInt size)
 {
     zmemcpy(dst, src, ALIGN_UP(size, 8) + sizeof(struct dfltcc_state));
 }
 
 static const int PAGE_ALIGN = 0x1000;
 
-voidpf ZLIB_INTERNAL dfltcc_alloc_window(strm, items, size)
-    z_streamp strm;
-    uInt items;
-    uInt size;
+voidpf ZLIB_INTERNAL dfltcc_alloc_window(z_streamp strm, uInt items, uInt size)
 {
     voidpf p, w;
 
     /* To simplify freeing, we store the pointer to the allocated buffer right
-     * before the window.
+     * before the window. Note that DFLTCC always uses HB_SIZE bytes.
      */
-    p = ZALLOC(strm, sizeof(voidpf) + items * size + PAGE_ALIGN,
+    p = ZALLOC(strm, sizeof(voidpf) + MAX(items * size, HB_SIZE) + PAGE_ALIGN,
                sizeof(unsigned char));
     if (p == NULL)
         return NULL;
-    w = ALIGN_UP((char FAR *)p + sizeof(voidpf), PAGE_ALIGN);
-    *(voidpf *)((char FAR *)w - sizeof(voidpf)) = p;
+    w = ALIGN_UP((char *)p + sizeof(voidpf), PAGE_ALIGN);
+    *(voidpf *)((char *)w - sizeof(voidpf)) = p;
     return w;
 }
 
-void ZLIB_INTERNAL dfltcc_free_window(strm, w)
-    z_streamp strm;
-    voidpf w;
+void ZLIB_INTERNAL dfltcc_copy_window(void *dest, const void *src, size_t n)
+{
+    memcpy(dest, src, MAX(n, HB_SIZE));
+}
+
+void ZLIB_INTERNAL dfltcc_free_window(z_streamp strm, voidpf w)
 {
     if (w)
-        ZFREE(strm, *(voidpf *)((unsigned char FAR *)w - sizeof(voidpf)));
+        ZFREE(strm, *(voidpf *)((unsigned char *)w - sizeof(voidpf)));
 }
 
 /*
@@ -858,15 +843,12 @@ void ZLIB_INTERNAL dfltcc_free_window(strm, w)
    fly with deflateParams, we need to convert between hardware and software
    window formats.
 */
-int ZLIB_INTERNAL dfltcc_deflate_params(strm, level, strategy, flush)
-    z_streamp strm;
-    int level;
-    int strategy;
-    int *flush;
+int ZLIB_INTERNAL dfltcc_deflate_params(z_streamp strm, int level,
+					int strategy, int *flush)
 {
-    deflate_state FAR *state = (deflate_state FAR *)strm->state;
-    struct dfltcc_state FAR *dfltcc_state = GET_DFLTCC_STATE(state);
-    struct dfltcc_param_v0 FAR *param = &dfltcc_state->param;
+    deflate_state *state = (deflate_state *)strm->state;
+    struct dfltcc_state *dfltcc_state = GET_DFLTCC_STATE(state);
+    struct dfltcc_param_v0 *param = &dfltcc_state->param;
     int could_deflate = dfltcc_can_deflate(strm);
     int can_deflate = dfltcc_can_deflate_with_params(strm,
                                                      level,
@@ -888,13 +870,11 @@ int ZLIB_INTERNAL dfltcc_deflate_params(strm, level, strategy, flush)
     return Z_OK;
 }
 
-int ZLIB_INTERNAL dfltcc_deflate_done(strm, flush)
-    z_streamp strm;
-    int flush;
+int ZLIB_INTERNAL dfltcc_deflate_done(z_streamp strm, int flush)
 {
-    deflate_state FAR *state = (deflate_state FAR *)strm->state;
-    struct dfltcc_state FAR *dfltcc_state = GET_DFLTCC_STATE(state);
-    struct dfltcc_param_v0 FAR *param = &dfltcc_state->param;
+    deflate_state *state = (deflate_state *)strm->state;
+    struct dfltcc_state *dfltcc_state = GET_DFLTCC_STATE(state);
+    struct dfltcc_param_v0 *param = &dfltcc_state->param;
 
     /* When deflate(Z_FULL_FLUSH) is called with small avail_out, it might
      * close the block without resetting the compression state. Detect this
@@ -913,15 +893,10 @@ int ZLIB_INTERNAL dfltcc_deflate_done(strm, flush)
 /*
    Preloading history.
 */
-local void append_history OF((struct dfltcc_param_v0 FAR *param,
-                              Bytef *history,
-                              const Bytef *buf,
-                              uInt count));
-local void append_history(param, history, buf, count)
-    struct dfltcc_param_v0 FAR *param;
-    Bytef *history;
-    const Bytef *buf;
-    uInt count;
+local void append_history(struct dfltcc_param_v0 *param,
+			  Bytef *history,
+			  const Bytef *buf,
+			  uInt count)
 {
     size_t offset;
     size_t n;
@@ -952,14 +927,27 @@ local void append_history(param, history, buf, count)
     }
 }
 
-int ZLIB_INTERNAL dfltcc_deflate_set_dictionary(strm, dictionary, dict_length)
-    z_streamp strm;
-    const Bytef *dictionary;
-    uInt dict_length;
+local void get_history(struct dfltcc_param_v0 *param,
+		       const Bytef *history,
+		       Bytef *buf)
 {
-    deflate_state FAR *state = (deflate_state FAR *)strm->state;
-    struct dfltcc_state FAR *dfltcc_state = GET_DFLTCC_STATE(state);
-    struct dfltcc_param_v0 FAR *param = &dfltcc_state->param;
+    if (param->ho + param->hl <= HB_SIZE)
+        /* Circular history buffer does not wrap - copy one chunk */
+        memcpy(buf, history + param->ho, param->hl);
+    else {
+        /* Circular history buffer wraps - copy two chunks */
+        memcpy(buf, history + param->ho, HB_SIZE - param->ho);
+        memcpy(buf + HB_SIZE - param->ho, history, param->ho + param->hl - HB_SIZE);
+    }
+}
+
+int ZLIB_INTERNAL dfltcc_deflate_set_dictionary(z_streamp strm,
+						const Bytef *dictionary,
+						uInt dict_length)
+{
+    deflate_state *state = (deflate_state *)strm->state;
+    struct dfltcc_state *dfltcc_state = GET_DFLTCC_STATE(state);
+    struct dfltcc_param_v0 *param = &dfltcc_state->param;
 
     append_history(param, state->window, dictionary, dict_length);
     state->strstart = 1; /* Add FDICT to zlib header */
@@ -967,29 +955,49 @@ int ZLIB_INTERNAL dfltcc_deflate_set_dictionary(strm, dictionary, dict_length)
     return Z_OK;
 }
 
-int ZLIB_INTERNAL dfltcc_deflate_get_dictionary(strm, dictionary, dict_length)
-    z_streamp strm;
-    Bytef *dictionary;
-    uInt *dict_length;
+int ZLIB_INTERNAL dfltcc_deflate_get_dictionary(z_streamp strm,
+						Bytef *dictionary,
+						uInt *dict_length)
 {
-    deflate_state FAR *state = (deflate_state FAR *)strm->state;
-    struct dfltcc_state FAR *dfltcc_state = GET_DFLTCC_STATE(state);
-    struct dfltcc_param_v0 FAR *param = &dfltcc_state->param;
+    deflate_state *state = (deflate_state *)strm->state;
+    struct dfltcc_state *dfltcc_state = GET_DFLTCC_STATE(state);
+    struct dfltcc_param_v0 *param = &dfltcc_state->param;
 
-    if (dictionary) {
-        if (param->ho + param->hl <= HB_SIZE)
-            /* Circular history buffer does not wrap - copy one chunk */
-            zmemcpy(dictionary, state->window + param->ho, param->hl);
-        else {
-            /* Circular history buffer wraps - copy two chunks */
-            zmemcpy(dictionary,
-                    state->window + param->ho,
-                    HB_SIZE - param->ho);
-            zmemcpy(dictionary + HB_SIZE - param->ho,
-                    state->window,
-                    param->ho + param->hl - HB_SIZE);
-        }
+    if (dictionary)
+        get_history(param, state->window, dictionary);
+    if (dict_length)
+        *dict_length = param->hl;
+    return Z_OK;
+}
+
+int ZLIB_INTERNAL dfltcc_inflate_set_dictionary(z_streamp strm,
+						const Bytef *dictionary,
+						uInt dict_length)
+{
+    struct inflate_state *state = (struct inflate_state *)strm->state;
+    struct dfltcc_state *dfltcc_state = GET_DFLTCC_STATE(state);
+    struct dfltcc_param_v0 *param = &dfltcc_state->param;
+
+    if (inflate_ensure_window(state)) {
+        state->mode = MEM;
+        return Z_MEM_ERROR;
     }
+
+    append_history(param, state->window, dictionary, dict_length);
+    state->havedict = 1;
+    return Z_OK;
+}
+
+int ZLIB_INTERNAL dfltcc_inflate_get_dictionary(z_streamp strm,
+						Bytef *dictionary,
+						uInt *dict_length)
+{
+    struct inflate_state *state = (struct inflate_state *)strm->state;
+    struct dfltcc_state *dfltcc_state = GET_DFLTCC_STATE(state);
+    struct dfltcc_param_v0 *param = &dfltcc_state->param;
+
+    if (dictionary && state->window)
+        get_history(param, state->window, dictionary);
     if (dict_length)
         *dict_length = param->hl;
     return Z_OK;
